unit Knowledge;

//  ////////////////////////////////////////////////////////////////////////////
//
//  Copyright (C) 2006 Jan Punter
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//  ////////////////////////////////////////////////////////////////////////////
//
//    lupd : 2006-03-31
//
//  ////////////////////////////////////////////////////////////////////////////

interface

uses

  Classes, SysUtils, DateUtils, MAth, Globals;

type

  EKnowledge = class( Exception);
  EKnParser  = class( EKnowledge);

  type TBits1  = 0 .. $0001;  const RA_MaskBits1  = High( TBits1 );
  type TBits2  = 0 .. $0003;  const RA_MaskBits2  = High( TBits2 );
  type TBits3  = 0 .. $0007;  const RA_MaskBits3  = High( TBits3 );
  type TBits4  = 0 .. $000f;  const RA_MaskBits4  = High( TBits4 );
  type TBits5  = 0 .. $001f;  const RA_MaskBits5  = High( TBits5 );
  type TBits6  = 0 .. $003f;  const RA_MaskBits6  = High( TBits6 );
  type TBits7  = 0 .. $007f;  const RA_MaskBits7  = High( TBits7 );
  type TBits8  = 0 .. $00ff;  const RA_MaskBits8  = High( TBits8 );
  type TBits9  = 0 .. $01ff;  const RA_MaskBits9  = High( TBits9 );
  type TBits10 = 0 .. $03ff;  const RA_MaskBits10 = High( TBits10);
  type TBits11 = 0 .. $07ff;  const RA_MaskBits11 = High( TBits11);
  type TBits12 = 0 .. $0fff;  const RA_MaskBits12 = High( TBits12);
  type TBits13 = 0 .. $1fff;  const RA_MaskBits13 = High( TBits13);
  type TBits14 = 0 .. $3fff;  const RA_MaskBits14 = High( TBits14);
  type TBits15 = 0 .. $7fff;  const RA_MaskBits15 = High( TBits15);
  type TBits16 = 0 .. $ffff;  const RA_MaskBits16 = High( TBits16);

const

  RA_Masks : array[ 0 .. 15] of Tbits16 = (
    RA_MaskBits1 ,
    RA_MaskBits2 ,
    RA_MaskBits3 ,
    RA_MaskBits4 ,
    RA_MaskBits5 ,
    RA_MaskBits6 ,
    RA_MaskBits7 ,
    RA_MaskBits8 ,
    RA_MaskBits9 ,
    RA_MaskBits10,
    RA_MaskBits11,
    RA_MaskBits12,
    RA_MaskBits13,
    RA_MaskBits14,
    RA_MaskBits15,
    RA_MaskBits16
  );


type

  TArray7      = array[ 0 .. 6] of Byte;
  TStringArray = array of string;
  TParamType   = ( ptUndefined, ptMode, ptParam);


  // ///////////////////////////////////////////////////////////////////////////

  TKnowledgeReader = class( TFileStream)
  private
    FFileName : string;
    FLine     : Integer;
    FColumn   : Integer;
    FInString : Boolean;
  public
    constructor Create( const aFileName: string);
    procedure   ParseError( const aMsg: string);
    function    ReadChar: Char;
    procedure   UnRead( anAmount : Integer);
    procedure   SkipWhiteSpace;
    procedure   ReadLParen;
    procedure   ReadRParen;
    function    ReadIdentifier: string;
    function    CheckIdentifier: Boolean;
    function    ReadInteger: Integer;
    function    ReadString: string;
    function    ReadStringArray: TStringArray;
    function    ReadTimeStamp: TDateTime;
    function    ReadIntSection        ( const aSectionName: string): Integer;
    function    ReadIdentifierSection ( const aSectionName: string): string;
    function    ReadStringSection     ( const aSectionName: string): string;
    function    ReadStringArraySection( const aSectionName: string): TStringArray;
    function    ReadTimeStampSection  ( const aSectionName: string): TDateTime;
  public
    property    Line   : Integer read FLine;
    property    Column : Integer read FColumn;
  end;


  // ///////////////////////////////////////////////////////////////////////////

  TParamKnowledge = class( TObject) // Used for both Params and Modes
  private
    FParamType    : TParamType;     // ptUndefined, ptMode or ptParam
    FRangeType    : string;         // Used for mapping names to values
    FLowValue     : Integer;        // Lowest  possible parameter value
    FHighValue    : Integer;        // Highest possible parameter value
    FDefaultValue : Integer;        // Default          parameter value
    FDefinitions  : TStringArray;   // More or less formal descriptions
    FComments     : string;         // Comments on definition
  private
    function    GetRange: string;
  public
    constructor Create(
      aParamType         : TParamType;
      const aRangeType   : string;
      aLowValue          : Integer;
      aHighValue         : Integer;
      aDefaultValue      : Integer;
      const aDefinitions : TStringArray;
      const aComments    : string
    );
  public
    property    ParamType    : TParamType   read FParamType   ;
    property    RangeType    : string       read FRangeType   ;
    property    LowValue     : Integer      read FLowValue    ;
    property    HighValue    : Integer      read FHighValue   ;
    property    DefaultValue : Integer      read FDefaultValue;
    property    Definitions  : TStringArray read FDefinitions ;
    property    Comments     : string       read FComments    ;
    property    Range        : string       read GetRange     ;
  end;


  // ///////////////////////////////////////////////////////////////////////////

  TParamKnowledgeTable = class( TObject)
  private
    FReader         : TKnowledgeReader;
    FParamKnowledge : array of TParamKnowledge;
    FTimeStamp      : TDateTime;
    FOsVersion      : string;
    FOsBuild        : Integer;
  private
    function    GetParamKnowledgeCount: Integer;
    procedure   SetParamKnowledgeCount( aValue: Integer);
    function    GetParamKnowledge( anIndex: Integer): TParamKnowledge;
    function    GetParamKnowledgeByType( const anIndex: string): TParamKnowledge;
  private
    procedure   ParseError( const aMsg: string);
    procedure   AddKnowledge( aParamKnowledge: TParamKnowledge);
  private
    function    ReadParamDef: Boolean;
    procedure   ReadParamKnowledge;
  private
    procedure   SwapItems( i, j: Integer);
    function    ItemCompare( i, j: Integer): Integer;
    procedure   QuickSort( L, R: Integer);
  protected
    procedure   Load;
    procedure   Sort;
  public
    constructor Create( const aFileName: string);
    destructor  Destroy;                                               override;
  public
    property    ParamKnowledgeCount: Integer    read GetParamKnowledgeCount;
    property    TimeStamp           : TDateTime read FTimeStamp;
    property    OSVersion           : string    read FOsVersion;
    property    OSBuild             : Integer   read FOsBuild;
  public
    property    ParamKnowledge      [ anIndex: Integer     ]: TParamKnowledge read GetParamKnowledge;
    property    ParamKnowledgeByType[ const anIndex: string]: TParamKnowledge read GetParamKnowledgeByType;
  end;


  // ///////////////////////////////////////////////////////////////////////////

  TModuleKnowledge = class( TObject)
  private
    FTypeId      : Integer;
    FPage        : string;
    FPageIndex   : Integer;
    FShortName   : string;
    FLongName    : string;
    FHeight      : Integer;
    FInputNames  : TStringArray;
    FOutputNames : TStringArray;
    FParamNames  : TStringArray;
    FModeNames   : TStringArray;
    FInputTypes  : TStringArray;
    FOutputTypes : TStringArray;
    FParamRanges : TStringArray;
    FModeRanges  : TStringArray;
    FParamTypes  : TStringArray;
    FModeTypes   : TStringArray;
  private
    procedure   AddInput     ( const aName: string);
    procedure   AddOutput    ( const aName: string);
    procedure   AddParam     ( const aName: string);
    procedure   AddMode      ( const aName: string);
    procedure   AddInputType ( const aName: string);
    procedure   AddOutputType( const aName: string);
    procedure   AddParamRange( const aName: string);
    procedure   AddModeRange ( const aName: string);
    procedure   AddParamType ( const aName: string);
    procedure   AddModeType  ( const aName: string);
  private
    function    GetInputCount     : Integer;
    function    GetOutputCount    : Integer;
    function    GetParamCount     : Integer;
    function    GetModeCount      : Integer;
    function    GetInputTypeCount : Integer;
    function    GetOutputTypeCount: Integer;
    function    GetParamRangeCount: Integer;
    function    GetModeRangeCount : Integer;
    function    GetParamTypeCount : Integer;
    function    GetModeTypeCount  : Integer;
    function    GetInputName ( anIndex: Integer): string;
    function    GetOutputName( anIndex: Integer): string;
    function    GetParamName ( anIndex: Integer): string;
    function    GetModeName  ( anIndex: Integer): string;
    function    GetInputType ( anIndex: Integer): string;
    function    GetOutputType( anIndex: Integer): string;
    function    GetParamRange( anIndex: Integer): string;
    function    GetModeRange ( anIndex: Integer): string;
    function    GetParamType ( anIndex: Integer): string;
    function    GetModeType  ( anIndex: Integer): string;
  public
    procedure   Learn(
            aTypeId       : Integer;
      const aPage         : string;
            aPageIndex    : Integer;
      const aShortName    : string;
      const aLongName     : string;
            aHeight       : Integer;
      const aInputs       : TStringArray;
      const aOutputs      : TStringArray;
      const aParams       : TStringArray;
      const aModes        : TStringArray;
      const anInputTypes  : TStringArray;
      const anOutputTypes : TStringArray;
      const aParamRanges  : TStringArray;
      const aModeRanges   : TStringArray;
      const aParamTypes   : TStringArray;
      const aModeTypes    : TStringArray
    );
    constructor Create;
    destructor  Destroy;                                               override;
    function    InputsAsString      : string;
    function    OutputsAsString     : string;
    function    ParamsAsString      : string;
    function    ModesAsString       : string;
    function    InputTypesAsString  : string;
    function    OutputTypesAsString : string;
    function    ParamRangesAsString : string;
    function    ModeRangesAsString  : string;
    function    ParamTypesAsString  : string;
    function    ModeTypesAsString   : string;
    procedure   DumpKnowledge( const aStrings: TStrings; anInd: Integer);
    procedure   DumpDetails  ( const aStrings: TStrings; anInd: Integer);
    function    ConnectorInfo( anIndex: Integer; isInput: Boolean): string;
  public
    property    TypeId          : Integer read FTypeId;
    property    Page            : string  read FPage;
    property    PageIndex       : Integer read FPageIndex;
    property    ShortName       : string  read FShortName;
    property    LongName        : string  read FLongName;
    property    Height          : Integer read FHeight;
    property    InputCount      : Integer read GetInputCount;
    property    OutputCount     : Integer read GetOutputCount;
    property    ParamCount      : Integer read GetParamCount;
    property    ModeCount       : Integer read GetModeCount;
    property    InputTypeCount  : Integer Read GetInputTypeCount;
    property    OutputTypeCount : Integer Read GetOutputTypeCount;
    property    ParamRangeCount : Integer Read GetParamRangeCount;
    property    ModeRangeCount  : Integer Read GetModeRangeCount;
    property    ParamTypeCount  : Integer Read GetParamTypeCount;
    property    ModeTypeCount   : Integer Read GetModeTypeCount;
  public
    property    InputName [ anIndex: Integer]: string read GetInputName;
    property    OutputName[ anIndex: Integer]: string read GetOutputName;
    property    ParamName [ anIndex: Integer]: string read GetparamName;
    property    ModeName  [ anIndex: Integer]: string read GetModeName;
    property    InputType [ anIndex: Integer]: string read GetInputType;
    property    OutputType[ anIndex: Integer]: string read GetOutputType;
    property    ParamRange[ anIndex: Integer]: string read GetParamRange;
    property    ModeRange [ anIndex: Integer]: string read GetModeRange;
    property    ParamType [ anIndex: Integer]: string read GetParamType;
    property    ModeType  [ anIndex: Integer]: string read GetModeType;
  end;


  // ///////////////////////////////////////////////////////////////////////////

  TModuleKnowledgeTable = class( TObject)
  private
    FReader          : TKnowledgeReader;
    FModuleKnowledge : array of TModuleKnowledge;
    FTimeStamp       : TDateTime;
    FOsVersion       : string;
    FOsBuild         : Integer;
  private
    function    GetModuleKnowledgeCount: Integer;
    procedure   SetModuleKnowledgeCount( aValue: Integer);
    function    GetKnowledge( anIndex: Integer): TModuleKnowledge;
    function    GetModuleKnowledge( aModuleId: Integer): TModuleKnowledge;
    function    GetPage           ( aModuleId: Integer): string;
    function    GetPageIndex      ( aModuleId: Integer): Integer;
    function    GetShortName      ( aModuleId: Integer): string;
    function    GetLongName       ( aModuleId: Integer): string;
    function    GetHeight         ( aModuleId: Integer): Integer;
    function    GetInputCount     ( aModuleId: Integer): Integer;
    function    GetOutputCount    ( aModuleId: Integer): Integer;
    function    GetParamCount     ( aModuleId: Integer): Integer;
    function    GetModeCount      ( aModuleId: Integer): Integer;
    function    GetInputName      ( aModuleId, anIndex: Integer): string;
    function    GetOutputName     ( aModuleId, anIndex: Integer): string;
    function    GetParamName      ( aModuleId, anIndex: Integer): string;
    function    GetModeName       ( aModuleId, anIndex: Integer): string;
  private
    procedure   ParseError( const aMsg: string);
    procedure   AddKnowledge( aModuleKnowledge: TModuleKnowledge);
  private
    function    ReadModuleDef: Boolean;
    procedure   ReadModuleKnowledge;
    procedure   Load;
  public
    constructor Create( const aFileName: string);
    destructor  Destroy;                                               override;
    procedure   DumpKnowledge( const aStringList: TStrings);           overload;
    procedure   DumpKnowledge( const aFileName: string);               overload;
  public
    property    ModuleKnowledgeCount: Integer read GetModuleKnowledgeCount;
  public
    property    TimeStamp : TDateTime read FTimeStamp;
    property    OsVersion : string    read FOsVersion;
    property    OsBuild   : integer   read FOsBuild;
  public
    property    Knowledge[ anIndex: Integer]: TModuleKnowledge read GetKnowledge;
  public
    property    ModuleKnowledge[ aModuleId: Integer]: TModuleKnowledge read GetModuleKnowledge;
    property    Page           [ aModuleId: Integer]: string           read GetPage;
    property    PageIndex      [ aModuleId: Integer]: Integer          read GetPageIndex;
    property    ShortName      [ aModuleId: Integer]: string           read GetShortName;
    property    LongName       [ aModuleId: Integer]: string           read GetLongName;
    property    Height         [ aModuleId: Integer]: Integer          read GetHeight;
    property    InputCount     [ aModuleId: Integer]: Integer          read GetInputCount;
    property    OutputCount    [ aModuleId: Integer]: Integer          read GetOutputCount;
    property    ParamCount     [ aModuleId: Integer]: Integer          read GetParamCount;
    property    ModeCount      [ aModuleId: Integer]: Integer          read GetModeCount;
  public
    property    InputName [ aModuleId, anIndex: Integer]: string read GetInputName;
    property    OutputName[ aModuleId, anIndex: Integer]: string read GetOutputName;
    property    ParamName [ aModuleId, anIndex: Integer]: string read GetParamName;
    property    ModeName  [ aModuleId, anIndex: Integer]: string read GetModeName;
  end;


  // ///////////////////////////////////////////////////////////////////////////

  TKnowledgePage = class( TObject)
  private
    FPageName  : string;
    FKnowledge : array of TModuleKnowledge;
  private
    function    GetKnowledgeCount: Integer;
    function    GetKnowledge( anIndex: Integer): TModuleKnowledge;
  public
    constructor Create( const aPageName: string);
    procedure   AddKnowledge( const aKnowledge: TModuleKnowledge);
  public
    property    PageName: string Read FPageName;
    property    KnowledgeCount: Integer read GetKnowledgeCount;
    property    Knowledge[ anIndex: Integer]: TModuleKnowledge read GetKnowledge;
  end;


  // ///////////////////////////////////////////////////////////////////////////

  TKnowledgePages = class( TObject)
  private
    FPages    : array of TKnowledgePage;
    FLockAdds : Boolean;
  private
    function    GetPageCount: Integer;
    function    GetKnowledgeCount( anIndex: Integer): Integer;
    function    GetPage( anIndex: Integer): TKnowledgePage;
    function    GetPageByName( const anIndex: string): TKnowledgePage;
    function    GetKnowledge( aPageIndex, aKnowledgeIndex: Integer): TModuleKnowledge;
    procedure   AddPage( const aPageName: string);
  public
    constructor Create;
    destructor  Destroy;                                               override;
    procedure   AddKnowledge( const aKnowledge: TModuleKnowledge);
  public
    property    PageCount                        : Integer read GetPageCount;
    property    KnowledgeCount[ anIndex: Integer]: Integer read GetKnowledgeCount;
  public
    property    Page      [ anIndex: Integer]                    : TKnowledgePage   read GetPage;
    property    PageByName[ const anIndex: string]               : TKnowledgePage   read GetPageByName;
    property    Knowledge [ aPageIndex, aKnowledgeIndex: Integer]: TModuleKnowledge read GetKnowledge;
  end;


  procedure CreatePagedKnowledge( aTable: TModuleKnowledgeTable);
  function  PagedKnowledge: TKnowledgePages;

  function  Bools( aVal: Boolean): string;
  function  PatchTypes( aVal: TBits1): string;
  function  MonoPolies( aVal: TBits2): string;
  function  CatName( aVal: TBits4): string;
  function  Locations( aVal: TBits2): string;
  function  CableTypes( aVal: TBits1): string;
  function  CableColors( aVal: TBits3): string;
  function  VariationName( aVal: TBits4): string;
  function  ByteToBinary( aByte, aLength: Byte): string;
  function  BytesToHex( const aBytes: array of byte): string;
  function  ParamTypeName( aParamType: TParamType): string;
  function  CrcClavia( Seed: Integer; aVal: Integer): Word;

  function  SplitParamRange( const aParamRange: string; var aLow, aHigh: Integer): Boolean;


implementation



  const

    Alpha        = [ 'A' .. 'Z', 'a' .. 'z'];
    Numeric      = [ '0' .. '9'];
    Specials     = [ '@', '_'];
    WhiteSpace   = [ ' ', #10, #13, ^I];
    AlphaNumeric = Alpha + Numeric;
    LeadChars    = Alpha + Specials;
    IdChars      = LeadChars + Numeric;



    ABools : array[ Boolean] of string = (
      'False',
      'True'
    );

    APatchTypes : array[ TBits1] of string = (
      'patch',
      'performance'
    );

    AMonoPolies : array[ TBits2] of string = (
      'poly',
      'mono',
      'legato',
      '<undefined_mono_poly_3>'
    );

    ACatName : array[ TBits4] of string = (
      'No Cat',
      'Acoustic',
      'Sequence',
      'Bass',
      'Classic',
      'Drum',
      'Fantasy',
      'FX',
      'Lead',
      'Organ',
      'Pad',
      'Piano',
      'Synth',
      'Audio In',
      'User 1',
      'User 2'
    );

    ALocations : array[ TBits2] of string = (
      'FX',
      'VA',
      '<undefined_area_2>',
      '<undefined_area_3>'
    );

    ACableTypes : array[ TBits1] of string = (
      'in-in',
      'out-in'
    );

    ACableColors : array[ TBits3] of string = (
      'Red',
      'Blue',
      'Yellow',
      'Orange',
      'Green',
      'Purple',
      'White',
      '<undefined_cable_color_7>'
    );

    AVariationName : array[ TBits4] of string = (
      '1',
      '2',
      '3',
      '4',
      '5',
      '6',
      '7',
      '8',
      'init',
      '<invalid_variation_09>',
      '<invalid_variation_10>',
      '<invalid_variation_11>',
      '<invalid_variation_12>',
      '<invalid_variation_13>',
      '<invalid_variation_14>',
      '<invalid_variation_15>'
    );

    AParamTypeName : array[ TParamType] of string = (
      '<undefined_param_type_name>',
      'Mode',
      'Param'
    );


    function Bools( aVal: Boolean): string;
    begin
      Result := ABools[ aVal];
    end;

    function PatchTypes( aVal: TBits1): string;
    begin
      Result := APatchTypes[ aVal];
    end;

    function MonoPolies( aVal: TBits2): string;
    begin
    end;

    function CatName( aVal: TBits4): string;
    begin
      Result := ACatName[ aVal];
    end;

    function Locations( aVal: TBits2): string;
    begin
      Result := ALocations[ aVal];
    end;

    function CableTypes( aVal: TBits1): string;
    begin
      Result := ACableTypes[ aVal];
    end;

    function CableColors( aVal: TBits3): string;
    begin
      Result := ACableColors[ aVal];
    end;

    function VariationName( aVal: TBits4): string;
    begin
      Result := AVariationName[ aVal];
    end;

    function ByteToBinary( aByte, aLength: Byte): string;
    begin
      Result := LongToBinary( aByte and RA_Masks[ aLength - 1], aLength);
    end;

    function BytesToHex( const aBytes: array of byte): string;
    var
      i : Integer;
    begin
      Result := '';
      for i := Low( aBytes) to High( aBytes)
      do Result := Result + Format( ' %.2x', [ aBytes[ i]]);
    end;

    function ParamTypeName( aParamType: TParamType): string;
    begin
      Result := AParamTypeName[ aParamType]
    end;

    function CrcClavia( Seed: Integer; aVal: Integer): Word;
    var
       i    : Integer;
       aCrc : Integer;
       k    : Integer;
    begin
       k    := ((( Seed shr 8) xor aVal) and 255) shl 8;
       aCrc := 0;
       for i := 1 to 8
       do begin
         if ( aCrc xor k) and $8000 <> 0
         then aCrc := ( aCrc shl 1) xor $1021
         else aCrc := aCrc shl 1;
         k := k shl 1;
       end;
       Result := (( Seed shl 8) xor aCrc) and $ffff;
    end;

    function  SplitParamRange( const aParamRange: string; var aLow, aHigh: Integer): Boolean;
    var
      N : String;
      p : Integer;
    begin
      aLow   :=  0;
      aHigh  := -1;
      Result := False;
      p      := 1;
      while ( p <= Length( aParamRange)) and ( aParamRange[ p] in WhiteSpace)
      do inc( p);
      if p <= Length( aParamRange)
      then begin
        N := '';
        while (( p <= Length( aParamRange)) and ( aParamRange[ p] in Numeric))
        do begin
          N := N + aParamRange[ p];
          Inc( p);
        end;
        if Length( N) > 0
        then begin
          aLow := StrToInt( N);
          while (( p <= Length( aParamRange)) and ( aParamRange[ p] in ( WhiteSpace + [ '.'])))
          do Inc( p);
          if p <= Length( aParamRange)
          then begin
            N := '';
            while (( p <= Length( aParamRange)) and ( aParamRange[ p] in Numeric))
            do begin
              N := N + aParamRange[ p];
              Inc( p);
            end;
            if Length( N) > 0
            then begin
              aHigh := StrToInt( N);
              Result := True;
            end;
          end;
        end;
      end;
    end;



var

  GPagedKnowledge : TKnowledgePages = nil;


  procedure CreatePagedKnowledge( aTable: TModuleKnowledgeTable);
  var
    i : Integer;
    K : TModuleKnowledge;
  begin
    if Assigned( aTable)
    then begin
      if Assigned( GPagedKnowledge)
      then FreeAndNil( GPagedKnowledge);
      GPagedKnowledge := TKnowledgePages.Create;
      with aTable
      do begin
        for i := 0 to ModuleKnowledgeCount - 1
        do begin
          K := Knowledge[ i];
          if Assigned( K)
          then GPagedKnowledge.AddKnowledge( K)
        end;
      end;
    end;
  end;


  procedure DestroyPagedKnowledge;
  begin
    if Assigned( GPagedKnowledge)
    then FreeAndNil( GPagedKnowledge);
  end;


  function  PagedKnowledge: TKnowledgePages;
  begin
    Result := GPagedKnowledge;
  end;



{ ========
  TKnowledgeReader = class( TFileStream)
  private
    FLine   : Integer;
    FColumn : Integer;
  public
    property    Line   : Integer read FLine;
    property    Column : Integer read FColumn;
  public
}

    constructor TKnowledgeReader.Create( const aFileName: string);
    begin
      inherited Create( aFileName, fmOpenRead + fmShareDenyWrite);
      FFileName := aFileName;
      FLine     := 1;
      FColumn   := 1;
    end;


    procedure   TKnowledgeReader.ParseError( const aMsg: string);
    begin
      Raise EKnParser.CreateFmt( 'Error at or near [%d,%d] : %s {file ''%s''}', [ FLine, FColumn, aMsg, FFileName]);
    end;


    function    TKnowledgeReader.ReadChar: Char;
    begin
      if Position < Size
      then begin
        Read( Result, SizeOf( Result));
        if Result = ^M   // CR
        then begin
          if FInString
          then ParseError( 'Unterminated string');
          Inc( FLine);
          FColumn := 1
        end
        else Inc( FColumn);
      end
      else ParseError( 'Unexpected end of file');
    end;


    procedure   TKnowledgeReader.UnRead( anAmount : Integer);
    var
      C : Char;
    begin
      // maybe todo : maybe make a pushback stack.
      //        the stack should keep track of Line/Col info
      while ( anAmount > 0) and ( Position > 0)
      do begin
        Dec( anAmount);
        Position := Position - 1;
        Read( C, SizeOf( C));
        if C = ^M
        then Dec( FLine);
        if FColumn > 1
        then Dec( FColumn);
        Position := Position - 1;
      end;
    end;


    procedure   TKnowledgeReader.SkipWhiteSpace;
    var
      C: Char;
    begin
      C := ReadChar;
      while C in WhiteSpace
      do C := ReadChar;
      UnRead( 1);
    end;


    procedure   TKnowledgeReader.ReadLParen;
    begin
      SkipWhiteSpace;
      if ReadChar <> '('
      then ParseError( '''('' expected');
    end;


    procedure   TKnowledgeReader.ReadRParen;
    begin
      SkipWhiteSpace;
      if ReadChar <> ')'
      then ParseError( '''('' expected');
    end;


    function    TKnowledgeReader.ReadIdentifier: string;
    var
      C: Char;
    begin
      Result := '';
      SkipWhiteSpace;
      C := ReadChar;
      if C in LeadChars
      then begin
        while C in IdChars
        do begin
          Result := Result + C;
          C := ReadChar;
        end;
        UnRead( 1);
      end
      else ParseError( 'Identifier expected');
    end;


    function    TKnowledgeReader.CheckIdentifier: Boolean;
    var
      C    : Char;
      aPos : Integer;
    begin
      Result := False;
      aPos   := Position;
      try
        SkipWhiteSpace;
        C := ReadChar;
        if C in LeadChars
        then Result := True;
      finally
        UnRead( Position - aPos);
      end;
    end;


    function    TKnowledgeReader.ReadInteger: Integer;
    var
      C : Char;
      N : string;
    begin
      Result := 0;
      SkipWhiteSpace;
      N := '';
      C := ReadChar;
      if C in Numeric
      then begin
        while C in Numeric
        do begin
          N := N + C;
          C := ReadChar;
        end;
        Unread( 1);
        Result := StrToInt( N);
      end
      else ParseError( 'Integer expected');
    end;


    function    TKnowledgeReader.ReadString: string;
    // read a string contained in single quotes, but accept
    // a double single quote ('') inside the string as
    // representing one (').
    var
      C    : Char;
      Done : Boolean;
    begin
      Result := '';
      SkipWhiteSpace;
      C := ReadChar;
      if C = ''''
      then begin
        Done      := False;
        FInString := True;
        try
          while not Done
          do begin
            C := ReadChar;
            while C <> ''''
            do begin
              Result := Result + C;
              C := ReadChar;
            end;
            C := ReadChar;
            if C <> ''''
            then begin
              Unread( 1);
              Done := true;
            end
            else Result := Result + C;
          end;
        finally
          FInString := False;
        end;
      end
      else ParseError( 'String expected');
    end;


    function    TKnowledgeReader.ReadStringArray: TStringArray;
    var
      C : Char;
      S : string;
    begin
      SetLength( Result, 0);
      SkipWhiteSpace;
      C := ReadChar;
      if C = '['
      then begin
        SkipWhiteSpace;
        C := ReadChar;
        while C <> ']'
        do begin
          Unread( 1);
          S := ReadString;
          SetLength( Result, Length( Result) + 1);
          Result[ Length( Result) - 1] := S;
          SkipWhiteSpace;
          C := ReadChar;
          if C = ','
          then begin
            SkipWhiteSpace;
            C := ReadChar;
          end;
        end;
      end
      else ParseError( 'String array expected');
    end;


    function    TKnowledgeReader.ReadTimeStamp: TDateTime;

      function StrReadInt( var S: string; var anInt: Integer): Boolean;
      var
        N : string;
      begin
        N := '';
        while ( Length( S) > 0) and ( not ( S[ 1] in Numeric))
        do S := Copy( S, 2, Length( S) - 1);
        while ( Length( S) > 0) and ( S[ 1] in Numeric)
        do begin
          N := N + S[ 1];
          S := Copy( S, 2, Length( S) - 1);
        end;
        if Length( N) > 0
        then begin
          anInt  := StrToInt( N);
          Result := True;
        end
        else Result := False;
      end;

    var
      S       : string;
      aYear   : Integer;
      aMonth  : Integer;
      aDay    : Integer;
      anHour  : Integer;
      aMinute : Integer;
    begin
      S := ReadString;
      if not StrReadInt( S, aYear)
      then ParseError( 'Integer expected (year in date/time)');
      if not StrReadInt( S, aMonth)
      then ParseError( 'Integer expected (month in date/time)');
      if not StrReadInt( S, aDay)
      then ParseError( 'Integer expected (day in date/time)');
      if not StrReadInt( S, anHour)
      then ParseError( 'Integer expected (hour in date/time)');
      if not StrReadInt( S, aMinute)
      then ParseError( 'Integer expected (minute in date/time)');
      Result := EncodeDateTime( aYear, aMonth, aDay, anHour, aMinute, 0, 0);
    end;


    function    TKnowledgeReader.ReadIntSection( const aSectionName: string): Integer;
    var
      S : string;
      N : Integer;
    begin
      Result := -1;
      S      := ReadIdentifier;
      if SameText( S, aSectionName)
      then begin
        ReadLParen;
        N := ReadInteger;
        ReadRParen;
        Result := N;
      end
      else ParseError( Format( 'IntSection ''%s( int)'' expected', [ aSectionName]));
    end;


    function    TKnowledgeReader.ReadIdentifierSection( const aSectionName: string): string;
    var
      S : string;
    begin
      SetLength( Result, 0);
      S := ReadIdentifier;
      if SameText( S, aSectionName)
      then begin
        ReadLParen;
        Result := ReadIdentifier;
        ReadRParen;
      end
      else ParseError( Format( 'Identifier section ''%s( ''string'')'' expected', [ aSectionName]));
    end;

    function    TKnowledgeReader.ReadStringSection( const aSectionName: string): string;
    var
      S : string;
    begin
      SetLength( Result, 0);
      S := ReadIdentifier;
      if SameText( S, aSectionName)
      then begin
        ReadLParen;
        Result := ReadString;
        ReadRParen;
      end
      else ParseError( Format( 'String section ''%s( ''string'')'' expected', [ aSectionName]));
    end;


    function    TKnowledgeReader.ReadStringArraySection( const aSectionName: string): TStringArray;
    var
      S : string;
    begin
      SetLength( Result, 0);
      S := ReadIdentifier;
      if SameText( S, aSectionName)
      then begin
        ReadLParen;
        Result := ReadStringArray;
        ReadRParen;
      end
      else ParseError( Format( 'String array section ''%s( [ ''string'', ''etc.''])'' expected', [ aSectionName]));
    end;


    function    TKnowledgeReader.ReadTimeStampSection( const aSectionName: string): TDateTime;
    var
      S : string;
    begin
      Result := 0;
      S      := ReadIdentifier;
      if SameText( S, aSectionName)
      then begin
        ReadLParen;
        Result := ReadTimeStamp;
        ReadRParen;
      end
      else ParseError( Format( 'Timestamp section ''%s( ''string'')'' expected', [ aSectionName]));
    end;



{ ========
  TParamKnowledge = class( TObject) // Used for both Params and Modes
  private
    FParamType    : TParamType;     // ptMode or ptParam
    FRangeType    : string;         // Used for mapping names to values
    FLowValue     : Integer;        // Lowest  possible parameter value
    FHighValue    : Integer;        // Highest possible parameter value
    FDefaultValue : Integer;        // Default          parameter value
    FDefinitions  : TStringArray;   // More or less formal descriptions
    FComments     : string;         // Comments on definition
  public
    property    ParamType    : TParamType   read FParamType   ;
    property    RangeType    : string       read FRangeType   ;
    property    LowValue     : Integer      read FLowValue    ;
    property    HighValue    : Integer      read FHighValue   ;
    property    DefaultValue : Integer      read FDefaultValue;
    property    Definitions  : TStringArray read FDefinitions ;
    property    Comments     : string       read FComments    ;
    property    Range        : string       read GetRange     ;
  private
}

    function    TParamKnowledge.GetRange: string;
    begin
      Result := Format( '%d .. %d', [ LowValue, HighValue]);
    end;


//  public


    constructor TParamKnowledge.Create(
      aParamType         : TParamType;
      const aRangeType   : string;
      aLowValue          : Integer;
      aHighValue         : Integer;
      aDefaultValue      : Integer;
      const aDefinitions : TStringArray;
      const aComments    : string
    );
    begin
      inherited Create;
      FParamType    := aParamType   ;
      FRangeType    := aRangeType   ;
      FLowValue     := aLowValue    ;
      FHighValue    := aHighValue   ;
      FDefaultValue := aDefaultValue;
      FDefinitions  := aDefinitions ;
      FComments     := aComments    ;
    end;



{ ========
  TParamKnowledgeTable = class( TObject)
  private
    FReader         : TKnowledgeReader;
    FParamKnowledge : array of TParamRange;
  public
    property    ParamKnowledgeCount: Integer read GetParamKnowledgeCount;
  public
    property    ParamKnowledge      [ anIndex: Integer     ]: TParamKnowledge read GetParamKnowledge;
    property    ParamKnowledgeByType[ const anIndex: string]: TParamKnowledge read GetParamKnowledgeByType;
  private
}

    function    TParamKnowledgeTable.GetParamKnowledgeCount: Integer;
    begin
      Result := Length( FParamKnowledge);
    end;


    procedure   TParamKnowledgeTable.SetParamKnowledgeCount( aValue: Integer);
    var
      i : Integer;
      K : TParamKnowledge;
    begin
      for i := aValue to ParamKnowledgeCount - 1
      do begin
        K := FParamKnowledge[ i];
        FParamKnowledge[ i] := nil;
        if Assigned( K)
        then K.Free;
      end;
      SetLength( FParamKnowledge, aValue);
    end;


    function    TParamKnowledgeTable.GetParamKnowledge( anIndex: Integer): TParamKnowledge;
    begin
      if ( anIndex >= 0) and ( anIndex < ParamKnowledgeCount)
      then Result := FParamKnowledge[ anIndex]
      else Result := nil;
    end;


    function    TParamKnowledgeTable.GetParamKnowledgeByType( const anIndex: string): TParamKnowledge;
    var
      i : Integer;
    begin
      Result := nil;
      for i := 0 to ParamKnowledgeCount - 1
      do begin
        if SameText( anIndex, ParamKnowledge[ i].RangeType)
        then begin
          Result := ParamKnowledge[ i];
          Break;
        end;
      end;
    end;


//  private


    procedure   TParamKnowledgeTable.ParseError( const aMsg: string);
    begin
      FReader.ParseError( aMsg);
    end;


    procedure   TParamKnowledgeTable.AddKnowledge( aParamKnowledge: TParamKnowledge);
    begin
      Assert( Assigned( aParamKnowledge), 'TParamKnowledgeTable.AddKnowledge: can''t add <nil> knowledge');
      SetParamKnowledgeCount( ParamKnowledgeCount + 1);
      FParamKnowledge[ ParamKnowledgeCount - 1] := aParamKnowledge;
    end;


//  private


    function    TParamKnowledgeTable.ReadParamDef: Boolean;
    var
      aPos          : Integer;
      S             : string;
      aParamType    : TParamType;
      aRangeType    : string;
      aLowValue     : Integer;
      aHighValue    : Integer;
      aDefaultValue : Integer;
      aDefinitions  : TStringArray;
      aComments     : string;
    begin
      Result := False;
      SetLength( aDefinitions , 0);
      if FReader.CheckIdentifier
      then begin
        aPos := FReader.Position;
        S    := FReader.ReadIdentifier;
        if SameText( S, 'ParamDef')
        then begin
          FReader.ReadLparen;
          S := FReader.ReadIdentifierSection( 'ParamType');
          if SameText( S, 'Param')
          then aParamType := ptParam
          else if SameText( S, 'Mode')
          then aParamType := ptMode
          else begin
            aParamType := ptUndefined;
            ParseError( '''Param'' or ''Mode'' expected');
          end;
          aRangeType    := FReader.ReadStringSection     ( 'RangeType'   );
          aLowValue     := FReader.ReadIntSection        ( 'LowValue'    );
          aHighValue    := FReader.ReadIntSection        ( 'HighValue'   );
          aDefaultValue := FReader.ReadIntSection        ( 'DefaultValue');
          aDefinitions  := FReader.ReadStringArraySection( 'Definitions' );
          aComments     := FReader.ReadStringSection     ( 'Comments'    );
          FReader.ReadRParen;

          AddKnowledge(
            TParamKnowledge.
              Create(
                aParamType,
                aRangeType,
                aLowValue,
                aHighValue,
                aDefaultValue,
                aDefinitions,
                aComments
              )
            );
          Result := True;
        end
        else FReader.UnRead( FReader.Position - aPos)
      end;
    end;


    procedure   TParamKnowledgeTable.ReadParamKnowledge;
    var
      S : string;
    begin
      S := FReader.ReadIdentifier;
      if SameText( S, 'ParamKnowledge')
      then begin
        FReader.ReadLParen;
        FTimeStamp := FReader.ReadTimeStampSection( 'TimeStamp');
        FOsVersion := FReader.ReadStringSection   ( 'OSVersion');
        FOSBuild   := FReader.ReadIntSection      ( 'OSBuild'  );
        while ReadParamDef
        do ;
        FReader.ReadRParen;
      end
      else ParseError( '''ParamKnowledge()'' expected');
    end;


//  private


    procedure   TParamKnowledgeTable.SwapItems( i, j: Integer);
    var
      t : TParamKnowledge;
    begin
      t                   := FParamKnowledge[ i];
      FParamKnowledge[ i] := FParamKnowledge[ j];
      FParamKnowledge[ j] := t;
    end;


    function    TParamKnowledgeTable.ItemCompare( i, j: Integer): Integer;
    var
      Si, Sj : string;
    begin
      Si := ParamKnowledge[ i].RangeType;
      Sj := ParamKnowledge[ j].RangeType;
      Result := AnsiStrIComp( PChar( Si), PChar( Sj));
    end;

    
    procedure   TParamKnowledgeTable.QuickSort( L, R: Integer);
    var
      I, J, P: Integer;
    begin
      repeat
        I := L;
        J := R;
        P := ( L + R) shr 1;
        repeat
          while ItemCompare( I, P) < 0
          do Inc( I);
          while ItemCompare( J, P) > 0
          do Dec( J);
          if I <= J
          then begin
            SwapItems( I, J);
            if P = I
            then P := J
            else if P = J
            then P := I;
            Inc( I);
            Dec( J);
          end;
        until I > J;
        if L < J
        then QuickSort( L, J);
        L := I;
      until I >= R;
    end;


//  protected


    procedure   TParamKnowledgeTable.Load;
    begin
      FReader.Position := 0;
      ReadParamKnowledge;
      Sort;
    end;


    procedure   TParamKnowledgeTable.Sort;
    begin
      QuickSort( 0, ParamKnowledgeCount - 1);
    end;
    

//  public


    constructor TParamKnowledgeTable.Create( const aFileName: string);
    begin
      inherited Create;
      FReader := TKnowledgeReader.Create( aFileName);
      try
        Load;
      finally
        FreeAndNil( FReader);
      end;
    end;


    destructor  TParamKnowledgeTable.Destroy; // override;
    begin
      if Assigned( FReader)
      then FReader.Free;
      SetParamKnowledgeCount( 0);
      inherited;
    end;



{ ========
  TModuleKnowledge = class( TModule)
  private
    FTypeId      : Integer;
    FPage        : string;
    FPageIndex   : Integer;
    FShortName   : string;
    FLongName    : string;
    FHeight      : Integer;
    FInputNames  : TStringArray;
    FOutputNames : TStringArray;
    FParamNames  : TStringArray;
    FModeNames   : TStringArray;
    FInputTypes  : TStringArray;
    FOutputTypes : TStringArray;
    FParamRanges : TStringArray;
    FModeRanges  : TStringArray;
    FParamTypes  : TStringArray;
    FModeTypes   : TStringArray;
  public
    property    TypeId          : Integer read FTypeId;
    property    Page            : string  read FPage;
    property    PageIndex       : Integer read FPageIndex;
    property    ShortName       : string  read FShortName;
    property    LongName        : string  read FLongName;
    property    Height          : Integer read FHeight;
    property    InputCount      : Integer read GetInputCount;
    property    OutputCount     : Integer read GetOutputCount;
    property    ParamCount      : Integer read GetParamCount;
    property    ModeCount       : Integer read GetModeCount;
    property    InputTypeCount  : Integer Read GetInputTypeCount;
    property    OutputTypeCount : Integer Read GetOutputTypeCount;
    property    ParamRangeCount : Integer Read GetParamRangeCount;
    property    ModeRangeCount  : Integer Read GetModeRangeCount;
    property    ParamTypeCount  : Integer Read GetParamTypeCount;
    property    ModeTypeCount   : Integer Read GetModeTypeCount;
  public
    property    InputName [ anIndex: Integer]: string read GetInputName;
    property    OutputName[ anIndex: Integer]: string read GetOutputName;
    property    ParamName [ anIndex: Integer]: string read GetparamName;
    property    ModeName  [ anIndex: Integer]: string read GetModeName;
    property    InputType [ anIndex: Integer]: string read GetInputType;
    property    OutputType[ anIndex: Integer]: string read GetOutputType;
    property    ParamRange[ anIndex: Integer]: string read GetParamRange;
    property    ModeRange [ anIndex: Integer]: string read GetModeRange;
    property    ParamType [ anIndex: Integer]: string read GetParamType;
    property    ModeType  [ anIndex: Integer]: string read GetModeType;
  private
}

    procedure   TModuleKnowledge.AddInput( const aName: string);
    begin
      SetLength( FInputNames, Length( FInputNames) + 1);
      FInputNames[ Length( FInputNames) - 1] := aName;
    end;


    procedure   TModuleKnowledge.AddOutput( const aName: string);
    begin
      SetLength( FOutputNames, Length( FOutputNames) + 1);
      FOutputNames[ Length( FOutputNames) - 1] := aName;
    end;


    procedure   TModuleKnowledge.AddParam( const aName: string);
    begin
      SetLength( FParamNames, Length( FParamNames) + 1);
      FParamNames[ Length( FParamNames) - 1] := aName;
    end;


    procedure   TModuleKnowledge.AddMode( const aName: string);
    begin
      SetLength( FModeNames, Length( FModeNames) + 1);
      FModeNames[ Length( FModeNames) - 1] := aName;
    end;


    procedure   TModuleKnowledge.AddInputType( const aName: string);
    begin
      SetLength( FInputTypes, Length( FInputTypes) + 1);
      FInputTypes[ Length( FInputTypes) - 1] := aName;
    end;


    procedure   TModuleKnowledge.AddOutputType( const aName: string);
    begin
      SetLength( FOutputTypes, Length( FOutputTypes) + 1);
      FOutputTypes[ Length( FOutputTypes) - 1] := aName;
    end;


    procedure   TModuleKnowledge.AddParamRange( const aName: string);
    begin
      SetLength( FParamRanges, Length( FParamRanges) + 1);
      FParamRanges[ Length( FParamRanges) - 1] := aName;
    end;


    procedure   TModuleKnowledge.AddModeRange( const aName: string);
    begin
      SetLength( FModeRanges, Length( FModeRanges) + 1);
      FModeRanges[ Length( FModeRanges) - 1] := aName;
    end;


    procedure   TModuleKnowledge.AddParamType( const aName: string);
    begin
      SetLength( FParamTypes, Length( FParamTypes) + 1);
      FParamTypes[ Length( FParamTypes) - 1] := aName;
    end;


    procedure   TModuleKnowledge.AddModeType( const aName: string);
    begin
      SetLength( FModeTypes, Length( FModeTypes) + 1);
      FModeTypes[ Length( FModeTypes) - 1] := aName;
    end;


//  private


    function    TModuleKnowledge.GetInputCount: Integer;
    begin
      Result := Length( FInputNames);
    end;


    function    TModuleKnowledge.GetOutputCount: Integer;
    begin
      Result := Length( FOutputNames);
    end;


    function    TModuleKnowledge.GetParamCount: Integer;
    begin
      Result := Length( FParamNames);
    end;


    function    TModuleKnowledge.GetModeCount: Integer;
    begin
      Result := Length( FModeNames);
    end;


    function    TModuleKnowledge.GetInputTypeCount: Integer;
    begin
      Result := Length( FInputTypes);
    end;


    function    TModuleKnowledge.GetOutputTypeCount: Integer;
    begin
      Result := Length( FOutputTypes);
    end;


    function    TModuleKnowledge.GetParamRangeCount: Integer;
    begin
      Result := Length( FParamRanges);
    end;


    function    TModuleKnowledge.GetModeRangeCount: Integer;
    begin
      Result := Length( FModeRanges);
    end;


    function    TModuleKnowledge.GetParamTypeCount: Integer;
    begin
      Result := Length( FParamTypes);
    end;


    function    TModuleKnowledge.GetModeTypeCount: Integer;
    begin
      Result := Length( FModeTypes);
    end;


    function    TModuleKnowledge.GetInputName( anIndex: Integer): string;
    begin
      if ( anIndex >= 0) and ( anIndex < InputCount)
      then Result := FInputNames[ anIndex]
      else Result := Format( '<Undefined Input: %d>', [ anIndex]);
    end;


    function    TModuleKnowledge.GetOutputName( anIndex: Integer): string;
    begin
      if ( anIndex >= 0) and ( anIndex < OutputCount)
      then Result := FOutputNames[ anIndex]
      else Result := Format( '<Undefined Output: %d>', [ anIndex]);
    end;


    function    TModuleKnowledge.GetParamName( anIndex: Integer): string;
    begin
      if ( anIndex >= 0) and ( anIndex < ParamCount)
      then Result := FParamNames[ anIndex]
      else Result := Format( '<Undefined Param: %d>', [ anIndex]);
    end;


    function    TModuleKnowledge.GetModeName( anIndex: Integer): string;
    begin
      if ( anIndex >= 0) and ( anIndex < ModeCount)
      then Result := FModeNames[ anIndex]
      else Result := Format( '<Undefined Mode: %d>', [ anIndex]);
    end;


    function    TModuleKnowledge.GetInputType( anIndex: Integer): string;
    begin
      if ( anIndex >= 0) and ( anIndex < InputTypeCount)
      then Result := FInputTypes[ anIndex]
      else Result := '<tbd>'
    end;


    function    TModuleKnowledge.GetOutputType( anIndex: Integer): string;
    begin
      if ( anIndex >= 0) and ( anIndex < OutputTypeCount)
      then Result := FOutputTypes[ anIndex]
      else Result := '<tbd>'
    end;


    function    TModuleKnowledge.GetParamRange( anIndex: Integer): string;
    begin
      if ( anIndex >= 0) and ( anIndex < ParamRangeCount)
      then Result := FParamRanges[ anIndex]
      else Result := '<tbd>'
    end;


    function    TModuleKnowledge.GetModeRange( anIndex: Integer): string;
    begin
      if ( anIndex >= 0) and ( anIndex < ModeRangeCount)
      then Result := FModeRanges[ anIndex]
      else Result := '<tbd>'
    end;


    function    TModuleKnowledge.GetParamType( anIndex: Integer): string;
    begin
      if ( anIndex >= 0) and ( anIndex < ParamTypeCount)
      then Result := FParamTypes[ anIndex]
      else Result := '<tbd>'
    end;


    function    TModuleKnowledge.GetModeType( anIndex: Integer): string;
    begin
      if ( anIndex >= 0) and ( anIndex < ModeTypeCount)
      then Result := FModeTypes[ anIndex]
      else Result := '<tbd>'
    end;


//  public


    procedure   TModuleKnowledge.Learn(
            aTypeId       : Integer;
      const aPage         : string;
            aPageIndex    : Integer;
      const aShortName    : string;
      const aLongName     : string;
            aHeight       : Integer;
      const aInputs       : TStringArray;
      const aOutputs      : TStringArray;
      const aParams       : TStringArray;
      const aModes        : TStringArray;
      const anInputTypes  : TStringArray;
      const anOutputTypes : TStringArray;
      const aParamRanges  : TStringArray;
      const aModeRanges   : TStringArray;
      const aParamTypes   : TStringArray;
      const aModeTypes    : TStringArray
    ); // virtual;
    var
      i : Integer;
    begin
      FTypeId    := aTypeId;
      FPage      := aPage;
      FPageIndex := aPageIndex;
      FShortName := aShortName;
      FLongName  := aLongName;
      FHeight    := aHeight;
      for i := Low( aInputs) to High( aInputs)
      do AddInput( aInputs[ i]);
      for i := Low( aOutputs) to High( aOutputs)
      do AddOutput( aOutputs[ i]);
      for i := Low( aParams) to High( aParams)
      do AddParam( aParams[ i]);
      for i := Low( aModes) to High( aModes)
      do AddMode( aModes[ i]);
      for i := Low( anInputTypes) to High( anInputTypes)
      do AddInputType( anInputTypes[ i]);
      for i := Low( anOutputTypes) to High( anOutputTypes)
      do AddOutputType( anOutputTypes[ i]);
      for i := Low( aParamRanges) to High( aParamRanges)
      do AddParamRange( aParamRanges[ i]);
      for i := Low( aModeRanges) to High( aModeRanges)
      do AddModeRange( aModeRanges[ i]);
      for i := Low( aParamTypes) to High( aParamTypes)
      do AddParamType( aParamTypes[ i]);
      for i := Low( aModeTypes) to High( aModeTypes)
      do AddModeType( aModeTypes[ i]);
    end;


    constructor TModuleKnowledge.Create;
    begin
      inherited;
      FPage      := '<invalid_page_name>';
      FShortName := '<invalid_short_name>';
      FLongName  := '<invalid_long_name>';
      FHeight    := 1;
    end;


    destructor  TModuleKnowledge.Destroy; // override;
    begin
      inherited;
    end;


    function    TModuleKnowledge.InputsAsString   : string;
    var
      i : Integer;
    begin
      Result := '[';
      for i := 0 to InputCount - 1
      do
        if i = 0
        then Result := Result + Format( ' ''%s''' , [ InputName[ i]])
        else Result := Result + Format( ', ''%s''', [ InputName[ i]]);
      Result := Result + ']';
    end;


    function    TModuleKnowledge.OutputsAsString : string;
    var
      i : Integer;
    begin
      Result := '[';
      for i := 0 to OutputCount - 1
      do
        if i = 0
        then Result := Result + Format( ' ''%s''' , [ OutputName[ i]])
        else Result := Result + Format( ', ''%s''', [ OutputName[ i]]);
      Result := Result + ']';
    end;


    function    TModuleKnowledge.ParamsAsString  : string;
    var
      i : Integer;
    begin
      Result := '[';
      for i := 0 to ParamCount - 1
      do
        if i = 0
        then Result := Result + Format( ' ''%s''' , [ ParamName[ i]])
        else Result := Result + Format( ', ''%s''', [ ParamName[ i]]);
      Result := Result + ']';
    end;


    function    TModuleKnowledge.ModesAsString   : string;
    var
      i : Integer;
    begin
      Result := '[';
      for i := 0 to ModeCount - 1
      do
        if i = 0
        then Result := Result + Format( ' ''%s''' , [ ModeName[ i]])
        else Result := Result + Format( ', ''%s''', [ ModeName[ i]]);
      Result := Result + ']';
    end;


    function    TModuleKnowledge.InputTypesAsString: string;
    var
      i : Integer;
    begin
      Result := '[';
      for i := 0 to InputTypeCount - 1
      do
        if i = 0
        then Result := Result + Format( ' ''%s''' , [ InputType[ i]])
        else Result := Result + Format( ', ''%s''', [ InputType[ i]]);
      Result := Result + ']';
    end;


    function    TModuleKnowledge.OutputTypesAsString: string;
    var
      i : Integer;
    begin
      Result := '[';
      for i := 0 to OutputTypeCount - 1
      do
        if i = 0
        then Result := Result + Format( ' ''%s''' , [ OutputType[ i]])
        else Result := Result + Format( ', ''%s''', [ OutputType[ i]]);
      Result := Result + ']';
    end;


    function    TModuleKnowledge.ParamRangesAsString: string;
    var
      i : Integer;
    begin
      Result := '[';
      for i := 0 to ParamRangeCount - 1
      do
        if i = 0
        then Result := Result + Format( ' ''%s''' , [ ParamRange[ i]])
        else Result := Result + Format( ', ''%s''', [ ParamRange[ i]]);
      Result := Result + ']';
    end;


    function    TModuleKnowledge.ModeRangesAsString: string;
    var
      i : Integer;
    begin
      Result := '[';
      for i := 0 to ModeRangeCount - 1
      do
        if i = 0
        then Result := Result + Format( ' ''%s''' , [ ModeRange[ i]])
        else Result := Result + Format( ', ''%s''', [ ModeRange[ i]]);
      Result := Result + ']';
    end;


    function    TModuleKnowledge.ParamTypesAsString: string;
    var
      i : Integer;
    begin
      Result := '[';
      for i := 0 to ParamTypeCount - 1
      do
        if i = 0
        then Result := Result + Format( ' ''%s''' , [ ParamType[ i]])
        else Result := Result + Format( ', ''%s''', [ ParamType[ i]]);
      Result := Result + ']';
    end;


    function    TModuleKnowledge.ModeTypesAsString: string;
    var
      i : Integer;
    begin
      Result := '[';
      for i := 0 to ModeTypeCount - 1
      do
        if i = 0
        then Result := Result + Format( ' ''%s''' , [ ModeType[ i]])
        else Result := Result + Format( ', ''%s''', [ ModeType[ i]]);
      Result := Result + ']';
    end;


    procedure   TModuleKnowledge.DumpKnowledge( const aStrings: TStrings; anInd: Integer);
    begin
      with aStrings
      do begin
        Add( IndFmt( anInd, 'ModuleDef('          , [                     ]));
        Add( IndFmt( anInd, '  ModuleType( %d)'   , [ TypeId              ]));
        Add( IndFmt( anInd, '  Page( ''%s'')'     , [ Page                ]));
        Add( IndFmt( anInd, '  PageIndex( %d)'    , [ PageIndex           ]));
        Add( IndFmt( anInd, '  ShortName( ''%s'')', [ ShortName           ]));
        Add( IndFmt( anInd, '  LongName( ''%s'')' , [ LongName            ]));
        Add( IndFmt( anInd, '  Height( %d)'       , [ Height              ]));
        Add( IndFmt( anInd, '  Inputs( %s)'       , [ InputsAsString      ]));
        Add( IndFmt( anInd, '  Outputs( %s)'      , [ OutputsAsString     ]));
        Add( IndFmt( anInd, '  Params( %s)'       , [ ParamsAsString      ]));
        Add( IndFmt( anInd, '  Modes( %s)'        , [ ModesAsString       ]));
        Add( IndFmt( anInd, '  InputTypes( %s)'   , [ InputTypesAsString  ]));
        Add( IndFmt( anInd, '  OutputTypes( %s)'  , [ OutputTypesAsString ]));
        Add( IndFmt( anInd, '  ParamRanges( %s)'  , [ ParamRangesAsString ]));
        Add( IndFmt( anInd, '  ModeRanges( %s)'   , [ ModeRangesAsString  ]));
        Add( IndFmt( anInd, '  ParamTypes( %s)'   , [ ParamTypesAsString  ]));
        Add( IndFmt( anInd, '  ModeTypes( %s)'    , [ ModeTypesAsString   ]));
        Add( IndFmt( anInd, ')             '      , [                     ]));
      end;
    end;


    procedure   TModuleKnowledge.DumpDetails( const aStrings: TStrings; anInd: Integer);
    begin
      with aStrings
      do begin
        Add( IndFmt( anInd, 'Knowledge['           , [                 ]));
        Add( IndFmt( anInd, '  Page       = ''%s''', [ Page            ]));
        Add( IndFmt( anInd, '  PageIndex  = %d'    , [ PageIndex       ]));
        Add( IndFmt( anInd, '  ShortName  = ''%s''', [ ShortName       ]));
        Add( IndFmt( anInd, '  LongName   = ''%s''', [ LongName        ]));
        Add( IndFmt( anInd, '  Height     = %d'    , [ Height          ]));
        Add( IndFmt( anInd, '  Inputs     = %s'    , [ InputsAsString  ]));
        Add( IndFmt( anInd, '  Outputs    = %s'    , [ OutputsAsString ]));
        Add( IndFmt( anInd, '  Params     = %s'    , [ ParamsAsString  ]));
        Add( IndFmt( anInd, '  Modes      = %s'    , [ ModesAsString   ]));
        Add( IndFmt( anInd, '] // Knowledge'       , [                 ]));
      end;
    end;


    function    TModuleKnowledge.ConnectorInfo( anIndex: Integer; isInput: Boolean): string;
    begin
      if isInput
      then begin
        if ( anIndex >= 0) and ( anIndex < InputCount)
        then Result := InputName[ anIndex]
        else Result := '<input_no_info>';
      end
      else begin
        if ( anIndex >= 0) and ( anIndex < OutputCount)
        then Result := OutputName[ anIndex]
        else Result := '<output_no_info>';
      end
    end;



{ ========
  TModuleKnowledgeTable = class( TObject)
  private
    FModuleKnowledge : array of TModuleKnowledge;
    FTimeStamp       : TDateTime;
    FOsVersion       : string;
    FOsBuild         : Integer;
  public
    property    TimeStamp : TDateTime read FTimeStamp;
    property    OsVersion : string    read FOsVersion;
    property    OsBuild   : integer   read FOsBuild;
  public
    property    ModuleKnowledgeCount: Integer read GetModuleKnowledgeCount;
  public
    property    Knowledge[ anIndex: Integer]: TModuleKnowledge read GetKnowledge;
  public
    property    ModuleKnowledge[ aModuleId: Integer]: TModuleKnowledge read GetModuleKnowledge;
    property    Page           [ aModuleId: Integer]: string           read GetPage;
    property    PageIndex      [ aModuleId: Integer]: Integer          read GetPageIndex;
    property    ShortName      [ aModuleId: Integer]: string           read GetShortName;
    property    LongName       [ aModuleId: Integer]: string           read GetLongName;
    property    Height         [ aModuleId: Integer]: Integer          read GetHeight;
    property    InputCount     [ aModuleId: Integer]: Integer          read GetInputCount;
    property    OutputCount    [ aModuleId: Integer]: Integer          read GetOutputCount;
    property    ParamCount     [ aModuleId: Integer]: Integer          read GetParamCount;
    property    ModeCount      [ aModuleId: Integer]: Integer          read GetModeCount;
  public
    property    InputName [ aModuleId, anIndex: Integer]: string read GetInputName;
    property    OutputName[ aModuleId, anIndex: Integer]: string read GetOutputName;
    property    ParamName [ aModuleId, anIndex: Integer]: string read GetParamName;
    property    ModeName  [ aModuleId, anIndex: Integer]: string read GetModeName;
  private
}

    function    TModuleKnowledgeTable.GetModuleKnowledgeCount: Integer;
    begin
      Result := Length( FModuleKnowledge);
    end;


    procedure   TModuleKnowledgeTable.SetModuleKnowledgeCount( aValue: Integer);
    var
      i : Integer;
      K : TModuleKnowledge;
    begin
      for i := aValue to ModuleKnowledgeCount - 1
      do begin
        K := FModuleKnowledge[ i];
        FModuleKnowledge[ i] := nil;
        if Assigned( K)
        then K.Free;
      end;
      SetLength( FModuleKnowledge, aValue);
    end;


    function    TModuleKnowledgeTable.GetKnowledge( anIndex: Integer): TModuleKnowledge;
    begin
      Assert(
        ( anIndex >= 0) and ( anIndex < ModuleKnowledgeCount),
        'TModuleKnowledgeTable.GetKnowledge : anIndex out of bounds'
      );
      Result := FModuleKnowledge[ anIndex];
    end;


    function    TModuleKnowledgeTable.GetModuleKnowledge( aModuleId: Integer): TModuleKnowledge;
    var
      i : Integer;
    begin
      Result := nil;
      for i := 0 to ModuleKnowledgeCount - 1
      do begin
        if FModuleKnowledge[ i].TypeId = aModuleId
        then begin
          Result := FModuleKnowledge[ i];
          Break;
        end;
      end;
    end;


    function    TModuleKnowledgeTable.GetPage( aModuleId: Integer): string;
    var
      K : TModuleKnowledge;
    begin
      K := ModuleKnowledge[ aModuleId];
      if Assigned( K)
      then Result := K.Page
      else Result := '<invalid_page>';
    end;


    function    TModuleKnowledgeTable.GetPageIndex( aModuleId: Integer): Integer;
    var
      K : TModuleKnowledge;
    begin
      K := ModuleKnowledge[ aModuleId];
      if Assigned( K)
      then Result := K.PageIndex
      else Result := -1;
    end;


    function    TModuleKnowledgeTable.GetShortName( aModuleId: Integer): string;
    var
      K : TModuleKnowledge;
    begin
      K := ModuleKnowledge[ aModuleId];
      if Assigned( K)
      then Result := K.ShortName
      else Result := '<invalid_short_name>';
    end;


    function    TModuleKnowledgeTable.GetLongName( aModuleId: Integer): string;
    var
      K : TModuleKnowledge;
    begin
      K := ModuleKnowledge[ aModuleId];
      if Assigned( K)
      then Result := K.LongName
      else Result := '<invalid_long_name>';
    end;


    function    TModuleKnowledgeTable.GetHeight( aModuleId: Integer): Integer;
    var
      K : TModuleKnowledge;
    begin
      K := ModuleKnowledge[ aModuleId];
      if Assigned( K)
      then Result := K.Height
      else Result := 1;
    end;


    function    TModuleKnowledgeTable.GetInputCount( aModuleId: Integer): Integer;
    var
      K : TModuleKnowledge;
    begin
      K := ModuleKnowledge[ aModuleId];
      if Assigned( K)
      then Result := K.InputCount
      else Result := 0;
    end;


    function    TModuleKnowledgeTable.GetOutputCount( aModuleId: Integer): Integer;
    var
      K : TModuleKnowledge;
    begin
      K := ModuleKnowledge[ aModuleId];
      if Assigned( K)
      then Result := K.OutputCount
      else Result := 0;
    end;


    function    TModuleKnowledgeTable.GetParamCount( aModuleId: Integer): Integer;
    var
      K : TModuleKnowledge;
    begin
      K := ModuleKnowledge[ aModuleId];
      if Assigned( K)
      then Result := K.ParamCount
      else Result := 0;
    end;


    function    TModuleKnowledgeTable.GetModeCount( aModuleId: Integer): Integer;
    var
      K : TModuleKnowledge;
    begin
      K := ModuleKnowledge[ aModuleId];
      if Assigned( K)
      then Result := K.ModeCount
      else Result := 0;
    end;


    function    TModuleKnowledgeTable.GetInputName( aModuleId, anIndex: Integer): string;
    var
      K : TModuleKnowledge;
    begin
      K := ModuleKnowledge[ aModuleId];
      if Assigned( K)
      then Result := K.InputName[ anIndex]
      else Result := '<invalid_input_name>';
    end;


    function    TModuleKnowledgeTable.GetOutputName( aModuleId, anIndex: Integer): string;
    var
      K : TModuleKnowledge;
    begin
      K := ModuleKnowledge[ aModuleId];
      if Assigned( K)
      then Result := K.OutputName[ anIndex]
      else Result := '<invalid_output_name>';
    end;


    function    TModuleKnowledgeTable.GetParamName( aModuleId, anIndex: Integer): string;
    var
      K : TModuleKnowledge;
    begin
      K := ModuleKnowledge[ aModuleId];
      if Assigned( K)
      then Result := K.ParamName[ anIndex]
      else Result := '<invalid_param_name>';
    end;


    function    TModuleKnowledgeTable.GetModeName( aModuleId, anIndex: Integer): string;
    var
      K : TModuleKnowledge;
    begin
      K := ModuleKnowledge[ aModuleId];
      if Assigned( K)
      then Result := K.ModeName[ anIndex]
      else Result := '<invalid_mode_name>';
    end;


//  private


    procedure   TModuleKnowledgeTable.ParseError( const aMsg: string);
    begin
      FReader.ParseError( aMsg);
    end;


    procedure   TModuleKnowledgeTable.AddKnowledge( aModuleKnowledge: TModuleKnowledge);
    begin
      Assert( Assigned( aModuleKnowledge), 'TModuleKnowledgeTable.AddKnowledge: can''t add <nil> knowledge');
      SetModuleKnowledgeCount( ModuleKnowledgeCount + 1);
      FModuleKnowledge[ ModuleKnowledgeCount - 1] := aModuleKnowledge;
    end;


//  private


    function    TModuleKnowledgeTable.ReadModuleDef: Boolean;
    var
      aPos          : Integer;
      S             : string;
      aKnowledge    : TModuleKnowledge;
      aModuleType   : Integer;
      aPage         : string;
      aPageIndex    : Integer;
      aShortName    : string;
      aLongName     : string;
      aHeight       : Integer;
      anInputs      : TStringArray;
      anOutputs     : TStringArray;
      aParams       : TStringArray;
      aModes        : TStringArray;
      anInputTypes  : TStringArray;
      anOutputTypes : TStringArray;
      aParamRanges  : TStringArray;
      aModeRanges   : TStringArray;
      aParamTypes   : TStringArray;
      aModeTypes    : TStringArray;
    begin
      Result := False;
      SetLength( anInputs     , 0);
      SetLength( anOutputs    , 0);
      SetLength( aParams      , 0);
      SetLength( aModes       , 0);
      SetLength( anInputTypes , 0);
      SetLength( anOutputTypes, 0);
      SetLength( aParamRanges , 0);
      SetLength( aModeRanges  , 0);
      SetLength( aParamTypes  , 0);
      SetLength( aModeTypes   , 0);
      if FReader.CheckIdentifier
      then begin
        aPos := FReader.Position;
        S    := FReader.ReadIdentifier;
        if SameText( S, 'ModuleDef')
        then begin
          aKnowledge := TModuleKnowledge.Create;
          try
            FReader.ReadLparen;
            aModuleType   := FReader.ReadIntSection        ( 'ModuleType' );
            aPage         := FReader.ReadStringSection     ( 'Page'       );
            aPageIndex    := FReader.ReadIntSection        ( 'PageIndex'  );
            aShortName    := FReader.ReadStringSection     ( 'ShortName'  );
            aLongName     := FReader.ReadStringSection     ( 'LongName'   );
            aHeight       := FReader.ReadIntSection        ( 'Height'     );
            anInputs      := FReader.ReadStringArraySection( 'Inputs'     );
            anOutputs     := FReader.ReadStringArraySection( 'Outputs'    );
            aParams       := FReader.ReadStringArraySection( 'Params'     );
            aModes        := FReader.ReadStringArraySection( 'Modes'      );
            anInputTypes  := FReader.ReadStringArraySection( 'InputTypes' );
            anOutputTypes := FReader.ReadStringArraySection( 'OutputTypes');
            aParamRanges  := FReader.ReadStringArraySection( 'ParamRanges');
            aModeRanges   := FReader.ReadStringArraySection( 'ModeRanges' );
            aParamTypes   := FReader.ReadStringArraySection( 'ParamTypes' );
            aModeTypes    := FReader.ReadStringArraySection( 'ModeTypes'  );
            FReader.ReadRParen;
            aKnowledge.
              Learn(
                aModuleType  ,
                aPage        ,
                aPageIndex   ,
                aShortName   ,
                aLongName    ,
                aHeight      ,
                anInputs     ,
                anOutputs    ,
                aParams      ,
                aModes       ,
                anInputTypes ,
                anOutputTypes,
                aParamRanges ,
                aModeRanges  ,
                aParamTypes  ,
                aModeTypes
              );
            AddKnowledge( aKnowledge);
            Result := True;
          except
            aKnowledge.Free;
            Raise;
          end;
        end
        else FReader.UnRead( FReader.Position - aPos)
      end;
    end;


    procedure   TModuleKnowledgeTable.ReadModuleKnowledge;
    var
      S : string;
    begin
      S := FReader.ReadIdentifier;
      if SameText( S, 'ModuleKnowledge')
      then begin
        FReader.ReadLParen;
        FTimeStamp := FReader.ReadTimeStampSection( 'TimeStamp');
        FOsVersion := FReader.ReadStringSection   ( 'OSVersion');
        FOSBuild   := FReader.ReadIntSection      ( 'OSBuild'  );
        while ReadModuleDef
        do ;
        FReader.ReadRParen;
      end
      else ParseError( '''ModuleKnowledge()'' expected');
    end;


    procedure   TModuleKnowledgeTable.Load;
    begin
      FReader.Position := 0;
      ReadModuleKnowledge;
    end;


//  public


    constructor TModuleKnowledgeTable.Create( const aFileName: string);
    begin
      inherited Create;
      FReader := TKnowledgeReader.Create( aFileName);
      try
        Load;
      finally
        FreeAndNil( FReader);
      end;
    end;


    destructor  TModuleKnowledgeTable.Destroy; // override;
    begin
      if Assigned( FReader)
      then FReader.Free;
      SetModuleKnowledgeCount( 0);
      inherited;
    end;


    procedure   TModuleKnowledgeTable.DumpKnowledge( const aStringList: TStrings); // overload;
    var
      i : Integer;
    begin
      with aStringList
      do begin
        Add( IndFmt( 0, 'ModuleKnowledge(', []));
        Add( IndFmt( 1,   'TimeStamp( ''%s'')', [ FormatDateTime( 'yyyy-mm-dd hh:nn', TimeStamp)]));
        Add( IndFmt( 1,   'OSVersion( ''%s'')', [ OsVersion]));
        Add( IndFmt( 1,   'OSBuild( %d)'      , [ OsBuild]));
        for i := 0 to ModuleKnowledgeCount - 1
        do Knowledge[ i].DumpKnowledge( aStringList, 1);
        Add( IndFmt( 0, ')', []));
      end;
    end;


    procedure   TModuleKnowledgeTable.DumpKnowledge( const aFileName: string); // overload;
    var
      aStringList: TStringList;
    begin
      aStringList := TStringList.Create;
      try
        with aStringList
        do begin
          Clear;
          DumpKnowledge( aStringList);
          SaveToFile( aFileName);
        end;
      finally
        aStringList.Free;
      end;
    end;



{ ========
  TKnowledgePage = class( TObject)
  private
    FPageName  : string;
    FKnowledge : array of TModuleKnowledge;
  public
    property    PageName: string Read FPageName;
    property    KnowledgeCount: Integer read GetKnowledgeCount;
    property    Knowledge[ anIndex: Integer]: TModuleKnowledge read GetKnowledge;
  private
}

    function    TKnowledgePage.GetKnowledgeCount: Integer;
    begin
      Result := Length( FKnowledge);
    end;


    function    TKnowledgePage.GetKnowledge( anIndex: Integer): TModuleKnowledge;
    begin
      Assert(
        ( anIndex >= 0) and ( anIndex < KnowledgeCount),
        Format( 'TKnowledgePage.GetModule : Invalid knowledge index (%d)', [ anIndex])
      );
      Result := FKnowledge[ anIndex];
      Assert(
        Assigned( Result),
        Format(
          'TKnowledgePage.GetModule : some knowledge (index %d) not present ...',
          [ anIndex]
        )
      );
    end;


//  public


    constructor TKnowledgePage.Create( const aPageName: string);
    begin
      inherited Create;
      FPageName := aPageName;
      SetLength( FKnowledge, 0);
    end;


    procedure   TKnowledgePage.AddKnowledge( const aKnowledge: TModuleKnowledge);
    var
      i : Integer;
    begin
      inherited;
      Assert(
        Assigned( aKnowledge),               // Can't add rubbish
        'TKnowledgePage.AddKnowledge : parameter ''aKnowledge'' not assigned'
      );
      Assert(
        FPageName = aKnowledge.Page,         // Iron out typo's in learned stuf
        'TKnowledgePage.AddKnowledge : ''aKnowledge'' about to be inserted into a wrong page'
      );
      for i := 0 to KnowledgeCount - 1
      do
        if
          Assigned( FKnowledge[ i]) and
          ( FKnowledge[ i].PageIndex = aKnowledge.PageIndex)
        then
          Assert(
            False,                        // Can't add twice with the same index
            Format(
              'TKnowledgePage.AddKnowledge : aKnowledge.PageIndex (%d) not unique',
              [ aKnowledge.PageIndex]
            )
          );
      if KnowledgeCount - 1 < aKnowledge.PageIndex
      then SetLength( FKnowledge, aKnowledge.PageIndex + 1);
      FKnowledge[ aKnowledge.PageIndex] := aKnowledge;
    end;



{ ========
  TKnowledgePages - class( TObject)
  private
    FPages : array of TKnowledgePage;
  public
    property    PageCount                     : Integer read GetPageCount;
    property    ModuleCount[ anIndex: Integer]: Integer read GetModuleCount;
  public
    property    Page      [ anIndex: Integer]                 : TKnowledgePage read GetPage;
    property    PageByName[ const anIndex: string]            : TKnowledgePage read GetPageByName;
    property    Module    [ aPageIndex, aModuleIndex: Integer]: TNoModule   read GetModule;
  private
}

    function    TKnowledgePages.GetPageCount: Integer;
    begin
      Result := Length( FPages);
    end;


    function    TKnowledgePages.GetKnowledgeCount( anIndex: Integer): Integer;
    begin
      Assert(
        ( anIndex >= 0) and ( anIndex < PageCount),
        Format(
          'TKnowledgePages.GetKnowledgeCount : invalid page index (%d)',
          [ anIndex]
        )
      );
      Assert(
        Assigned( FPages[ anIndex]),
        Format(
          'TKnowledgePages.GetKnowledgeCount : page (index %d) not present',
          [ anIndex]
        )
      );
      Result := FPages[ anIndex].KnowledgeCount;
    end;


    function    TKnowledgePages.GetPage( anIndex: Integer): TKnowledgePage;
    begin
      Assert(
        ( anIndex >= 0) and ( anIndex < PageCount),
        Format(
          'TKnowledgePages.GetModuleCount : invalid page index (%d)',
          [ anIndex]
        )
      );
      Result := FPages[ anIndex];
      Assert(
        Assigned( Result),
        Format(
          'TKnowledgePages.GetModuleCount : page (index %d) not present or <nil>',
          [ anIndex]
        )
      );
    end;


    function    TKnowledgePages.GetPageByName( const anIndex: string): TKnowledgePage;
    var
      i : Integer;
    begin
      Result := nil;
      for i := 0 to PageCount - 1
      do
        if SameText( Page[ i].FPageName, anIndex)
        then begin
          Result := Page[ i];
          Break;
        end;
      Assert(
        Assigned( Result),
        Format(
          'TKnowledgePages.GetModuleCount : some page (name ''%s'') not present',
          [ anIndex]
        )
      );
    end;


    function    TKnowledgePages.GetKnowledge( aPageIndex, aKnowledgeIndex: Integer): TModuleKnowledge;
    begin
      Result := FPages[ aPageIndex].Knowledge[ aKnowledgeIndex];
    end;


    procedure   TKnowledgePages.AddPage( const aPageName: string);
    begin
      Assert( not FLockAdds, 'TKnowledgePages.AddPage : Add''s are locked');
      SetLength( FPages, Length( FPages) + 1);
      FPages[ PageCount - 1] := TKnowledgePage.Create( aPageName);
    end;


//  public


    constructor TKnowledgePages.Create;
    begin
      inherited;
      FLockAdds := False;  // Allow for new pages
      AddPage( 'In/Out');
      AddPage( 'Note');
      AddPage( 'Osc');
      AddPage( 'LFO');
      AddPage( 'Rnd');
      AddPage( 'Env');
      AddPage( 'Filter');
      addpage( 'FX');
      AddPage( 'Delay');
      AddPage( 'Shaper');
      AddPage( 'Level');
      AddPage( 'Mixer');
      AddPage( 'Switch');
      AddPage( 'Logic');
      AddPage( 'Seq');
      AddPage( 'MIDI');
      FLockAdds := True;   // No more new pafge allowed
    end;


    destructor  TKnowledgePages.Destroy; // override;
    var
      i : Integer;
    begin
      for i := 0 to PageCount - 1
      do Page[ i].Free;
      inherited;
    end;


    procedure   TKnowledgePages.AddKnowledge( const aKnowledge: TModuleKnowledge);
    begin
      Assert(
        Assigned( aKnowledge),
        'TKnowledgePages.AddKnowledge : can''t add <nil> knowledge'
      );
      with aKnowledge
      do PageByName[ Page].AddKnowledge( aKnowledge);
    end;

    

initialization

  // Nothing, KnowledgeTable must be created explicitely
  // to avoid possible weird crashes at program startup.

finalization

  DestroyPagedKnowledge;

end.
