unit BitHose;

//  ////////////////////////////////////////////////////////////////////////////
//
//  Copyright (C) 2004 .. 2006 Jan Punter
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//  ////////////////////////////////////////////////////////////////////////////
//
//    lupd : 2006-03-26
//
//  ////////////////////////////////////////////////////////////////////////////


interface

uses

  Math;

type

  TBitCell   = type Cardinal;            // Cardinals used for bit cells - strong type checking
  TByteArray = array of Byte;            // Array of byte with a variable number of elements

const

  CellSize = 8 * SizeOf( TBitCell);      // Nr of bits in  each cell
  AllOnes  = TBitCell( -1);              // A TBitCell with all bits set to 1

type

  TBitIndex = 0 .. CellSize - 1;         // Range for bit indices
  TBitCount = 0 .. CellSize;             // Range for bit counts
  TBitValue = 0 .. 1;                    // Possible values for a single bit.

  TBitHose = class
  private
    FData     : array of TBitCell;
    FSize     : Integer;
    FWritePtr : Integer;
    FReadPtr  : Integer;
  private
    procedure   SetSize( aValue: Integer);
    function    GetBits( anIndex: Integer; aBitCount: TBitCount): TBitCell;
    procedure   SetBits( anIndex: Integer; aBitCount: TBitCount; aValue: TBitCell);
  public
    constructor Create;
    destructor  Destroy;                                               override;
    procedure   Write( aBitCount: TBitCount; aValue: TBitCell);
    procedure   ShiftInSeptets( const aData: TByteArray);
    procedure   ShiftInOctets ( const aData: TByteArray);
    procedure   AddSeptets( const aData: TByteArray);
    procedure   AddOctets ( const aData: TByteArray);
    procedure   Reset;
    procedure   Clear;
    function    GetOctets : TByteArray;                                overload;
    function    GetOctets( anOctetCount: Integer) : TByteArray;        overload;
    function    GetSeptets: TByteArray;
    function    Read( anAmount: TBitCount): TBitCell;                  overload;
    procedure   Read( var aData: TByteArray; anAmount: Integer);       overload;
    procedure   CopyTo  ( aBitHose: TBitHose; aStart, aBitCount: Integer);
    procedure   CopyFrom( aBitHose: TBitHose; aStart, aBitCount: Integer);
  public
    property    Size       : Integer read FSize    write SetSize;
    property    ReadPointer: Integer read FReadPtr write FReadPtr;
  public
    property    Bits[ aBitIndex: Integer; aBitCount: TBitCount]: TBitCell read GetBits write SetBits; default;
  end;


implementation


{ ========
  TBitHose = class
  private
    FData     : array of TBitCell;
    FSize     : Integer;
    FWritePtr : Integer;
    FReadPtr  : Integer;
  public
    property    Size       : Integer read FSize    write SetSize;
    property    ReadPointer: Integer read FReadPtr write FReadPtr;
  public
    property    Bits[ aBitIndex: Integer; aBitCount: TBitCount]: TBitCell read GetBits write SetBits; default;
  private
}

    procedure   TBitHose.SetSize( aValue: Integer);
    begin
      SetLength( FData, ( aValue + CellSize - 1 ) div CellSize);
      FSize := aValue;
    end;

    function    TBitHose.GetBits( anIndex: Integer; aBitCount: TBitCount): TBitCell;
    var
      CellPtr : Integer;
      BitPtr  : TBitIndex;
      Filled  : TBitCount;   // Nr of bits available in byte pointed to by CellPtr
      Shift   : Integer;     // >= 0 -> right shift,  < 0 -> left shift
    begin
      if aBitCount > 0
      then begin
        CellPtr := anIndex div CellSize;
        BitPtr  := anIndex mod CellSize;
        Filled  := CellSize - BitPtr;
        Shift   := Filled - aBitCount;
        if Shift >= 0
        then Result := ( FData[ CellPtr] Shr Shift) and ( AllOnes shr ( CellSize - aBitCount))
        else
          Result :=
            (( FData[ CellPtr] and ( AllOnes shr BitPtr)) shl -Shift) or
            ( FData[ CellPtr + 1] shr ( CellSize + Shift));
      end
      else Result := 0;
    end;

    procedure   TBitHose.SetBits( anIndex: Integer; aBitCount: TBitCount; aValue: TBitCell);
    var
      CellPtr : Integer;
      BitPtr  : TBitIndex;
      Free    : TBitCount;  // Free bits in cell pointed to by CellPtr
      Shift   : Integer;    // >= 0 -> left shift, < 0 -> right shift
    begin
      if aBitCount > 0
      then Begin
        CellPtr := anIndex div CellSize;
        BitPtr  := anIndex mod CellSize;
        Size    := Max( anIndex + aBitCount, Size); // Ensure the new bits will fit.
        Free    := CellSize - BitPtr;               //   so we indeed always have 1 .. 32 free bits ...
        Shift   := Free - aBitCount;
        aValue  := aValue and ( AllOnes shr ( CellSize - aBitCount));
        if Shift >= 0
        then FData[ CellPtr] := ( FData[ CellPtr] and ( AllOnes shl Free)) or ( aValue shl Shift)
        else begin
          FData[ CellPtr    ] := ( FData[ CellPtr] and ( AllOnes shl Free)) or ( aValue shr -Shift);
          FData[ CellPtr + 1] := aValue shl ( CellSize + Shift);
        end;
      end;
    end;

//   Public

    constructor TBitHose.Create;
    begin
      inherited;
      Clear;
    end;

    destructor  TBitHose.Destroy; // Override;
    begin
      Clear;
      inherited;
    end;

    procedure   TBitHose.Write( aBitCount: TBitCount; aValue: TBitCell);
    begin
      Bits[ FWritePtr, aBitCount] := aValue;
      Inc( FWritePtr, aBitCount);
    end;

    procedure   TBitHose.ShiftInSeptets( const aData: TByteArray);
    var
      i : Integer;
    begin
      for i := 0 to Length( aData) - 1
      do Write( 7, aData[ i]);
    end;

    procedure   TBitHose.ShiftInOctets( const aData: TByteArray);
    var
      i : Integer;
    begin
      for i := 0 to Length( aData) - 1
      do Write( 8, aData[ i]);
    end;

    procedure   TBitHose.AddSeptets( const aData: TByteArray);
    begin
      ShiftInSeptets( aData);
    end;

    procedure   TBitHose.AddOctets ( const aData: TByteArray);
    begin
      ShiftInOctets( aData);
    end;

    procedure   TBitHose.Reset;
    begin
      FReadPtr  := 0;
      FWritePtr := 0;
    end;

    procedure   TBitHose.Clear;
    begin
      Reset;
      Size := 0;
    end;

    function    TBitHose.GetOctets: TByteArray; // overload;
    var
      OldPtr : Integer;
    begin
      OldPtr   := FReadPtr;
      FReadPtr := 0;
      Result   := GetOctets( Size div 8);
      FReadPtr := OldPtr;
    end;

    function    TBitHose.GetOctets( anOctetCount: Integer) : TByteArray; // overload;
    var
      i : Integer;
    begin
      SetLength( Result, anOctetCount);
      for i := 0 to Length( Result) - 1
      do Result[ i] := Bits[ ReadPointer + i * 8, 8] and $ff;
    end;

    function    TBitHose.GetSeptets: TByteArray;
    var
      i : Integer;
    begin
      SetLength( Result, ( Size + 6) div 7);
      for i := 0 To Length( Result) - 1
      do Result[ i] := Bits[ i * 7, 7] and $7f;
    end;

    function    TBitHose.Read( anAmount: TBitCount): TBitCell;
    begin
      Result := Bits[ FReadPtr, anAmount];
      Inc( FReadPtr, anAmount);
    end;

    procedure   TBitHose.Read( var aData: TByteArray; anAmount: Integer); // overload;
    var
      i : Integer;
    begin
      SetLength( aData, 0);
      i := 0;
      while anAmount > 0
      do begin
        SetLength( aData, Length( aData) + 1);
        if anAmount > 8
        then aData[ i] := Read( 8)
        else aData[ i] := Read( anAmount);
        Dec( anAmount, 8);
        Inc( i);
      end;
    end;

    procedure   TBitHose.CopyTo( aBitHose: TBitHose; aStart, aBitCount: Integer);
    var
      i : Integer;
    begin
      if Assigned( aBitHose)
      then begin
        aBitHose.Size := 0;
        aBitHose.Reset;
        for i := 0 to aBitCount - 1
        do aBitHose.Bits[ i, 1] := Bits[ aStart + i, 1];
      end;
    end;

    procedure   TBitHose.CopyFrom( aBitHose: TBitHose; aStart, aBitCount: Integer);
    begin
      if Assigned( aBitHose)
      then aBitHose.CopyTo( Self, aStart, aBitCount);
    end;



end.
